from numbers import Real
from typing import (List,
                    Tuple)

Coordinate = Real
Point = Tuple[Coordinate, Coordinate]
BoundingBox = Tuple[Coordinate, Coordinate, Coordinate, Coordinate]
Segment = Tuple[Point, Point]
Contour = List[Point]



from reprlib import recursive_repr
from typing import Optional

from reprit.base import generate_repr
from robust.linear import segments_intersection

from orient.hints import (Coordinate,
                          Point,
                          Segment)


class Event:
    __slots__ = ('is_left_endpoint', 'start', 'complement',
                 'from_left_contour', 'below_from_right_contour_in_out')

    def __init__(self,
                 is_left_endpoint: bool,
                 start: Point,
                 complement: Optional['Event'],
                 from_left_contour: bool,
                 below_from_right_contour_in_out: Optional[bool] = None
                 ) -> None:
        self.is_left_endpoint = is_left_endpoint
        self.start = start
        self.complement = complement
        self.from_left_contour = from_left_contour
        self.below_from_right_contour_in_out = below_from_right_contour_in_out

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def end(self) -> Point:
        return self.complement.start

    @property
    def is_vertical(self) -> bool:
        start_x, _ = self.start
        end_x, _ = self.end
        return start_x == end_x

    @property
    def is_horizontal(self) -> bool:
        _, start_y = self.start
        _, end_y = self.end
        return start_y == end_y

    @property
    def segment(self) -> Segment:
        return self.start, self.end

    def y_at(self, x: Coordinate) -> Coordinate:
        if self.is_vertical or self.is_horizontal:
            _, start_y = self.start
            return start_y
        else:
            start_x, start_y = self.start
            if x == start_x:
                return start_y
            end_x, end_y = self.end
            if x == end_x:
                return end_y
            _, result = segments_intersection(self.segment,
                                              ((x, start_y), (x, end_y)))
            return result



from typing import Iterable

from orient.hints import (BoundingBox,
                          Contour,
                          Segment)


def contour_to_segments(contour: Contour) -> Iterable[Segment]:
    return ((contour[index - 1], contour[index])
            for index in range(len(contour)))


def contour_to_bounding_box(contour: Contour) -> BoundingBox:
    iterator = iter(contour)
    x_min, y_min = x_max, y_max = next(iterator)
    for x, y in iterator:
        x_min, x_max = min(x_min, x), max(x_max, x)
        y_min, y_max = min(y_min, y), max(y_max, y)
    return x_min, x_max, y_min, y_max



from robust.linear import (SegmentsRelationship,
                           segments_intersection,
                           segments_relationship)

from orient.hints import (Contour,
                          Point,
                          Segment)
from .event import Event
from .events_queue import (EventsQueue,
                           EventsQueueKey)
from .sweep_line import SweepLine
from .utils import (contour_to_bounding_box,
                    contour_to_segments)


def sweep(left: Contour,
          right: Contour) -> bool:
    events_queue = EventsQueue()
    for segment in contour_to_segments(left):
        register_segment(segment, True, events_queue)
    for segment in contour_to_segments(right):
        register_segment(segment, False, events_queue)
    sweep_line = SweepLine()
    while events_queue:
        event = events_queue.pop()
        start_x, _ = event.start
        sweep_line.move_to(start_x)
        if event.is_left_endpoint:
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            if below_event is None:
                event.below_from_right_contour_in_out = event.from_left_contour
            else:
                compute_transition(below_event, event, events_queue)
            if above_event is not None:
                compute_transition(event, above_event, events_queue)
            if (event.from_left_contour
                    and event.below_from_right_contour_in_out):
                return False
        else:
            event = event.complement
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    detect_intersection(below_event, above_event, events_queue)
    return True


def compute_transition(below_event: Event,
                       event: Event,
                       events_queue: EventsQueue) -> None:
    if detect_intersection(below_event, event, events_queue) == 2:
        if event.from_left_contour:
            event.below_from_right_contour_in_out = (
                below_event.below_from_right_contour_in_out)
        else:
            event.below_from_right_contour_in_out = (
                not below_event.below_from_right_contour_in_out)
    elif event.from_left_contour:
        event.below_from_right_contour_in_out = (
            below_event.below_from_right_contour_in_out)
    else:
        event.below_from_right_contour_in_out = (
            not below_event.below_from_right_contour_in_out)


def register_segment(segment: Segment,
                     from_left_contour: bool,
                     events_queue: EventsQueue) -> None:
    start, end = sorted(segment)
    start_event = Event(True, start, None, from_left_contour)
    end_event = Event(False, end, start_event, from_left_contour)
    start_event.complement = end_event
    events_queue.push(start_event)
    events_queue.push(end_event)


def detect_intersection(event: Event,
                        above_event: Event,
                        events_queue: EventsQueue) -> int:
    segment, above_segment = event.segment, above_event.segment
    relationship = segments_relationship(segment, above_segment)
    if relationship is SegmentsRelationship.NONE:
        # no intersection
        return 0
    elif relationship is SegmentsRelationship.CROSS:
        # segments intersect
        if (event.start == above_event.start
                or event.end == above_event.end):
            # segments intersect at an endpoint of both line segments
            return 0
        point = segments_intersection(segment, above_segment)
        if event.start != point and event.end != point:
            divide_segment(event, point, events_queue)
        if above_event.start != point and above_event.end != point:
            divide_segment(above_event, point, events_queue)
        return 1
    # segments overlap
    if event.from_left_contour is above_event.from_left_contour:
        raise ValueError('Edges of the same polygon should not overlap.')

    sorted_events = []
    starts_equal = event.start == above_event.start
    if starts_equal:
        sorted_events.append(None)
    elif EventsQueueKey(event) > EventsQueueKey(above_event):
        sorted_events.append(above_event)
        sorted_events.append(event)
    else:
        sorted_events.append(event)
        sorted_events.append(above_event)

    ends_equal = event.end == above_event.end
    if ends_equal:
        sorted_events.append(None)
    elif (EventsQueueKey(event.complement)
          > EventsQueueKey(above_event.complement)):
        sorted_events.append(above_event.complement)
        sorted_events.append(event.complement)
    else:
        sorted_events.append(event.complement)
        sorted_events.append(above_event.complement)

    if starts_equal:
        # both line segments are equal or share the left endpoint
        if not ends_equal:
            divide_segment(sorted_events[2].complement, sorted_events[1].start,
                           events_queue)
        return 2
    elif ends_equal:
        # the line segments share the right endpoint
        divide_segment(sorted_events[0], sorted_events[1].start, events_queue)
        return 3
    else:
        divide_segment(sorted_events[0]
                       # one line segment includes the other one
                       if sorted_events[0] is sorted_events[3].complement
                       # no line segment includes the other one
                       else sorted_events[1],
                       sorted_events[2].start,
                       events_queue)
        divide_segment(sorted_events[0], sorted_events[1].start, events_queue)
        return 3


def divide_segment(event: Event,
                   point: Point,
                   events_queue: EventsQueue) -> None:
    left_event = Event(True, point, event.complement, event.from_left_contour)
    right_event = Event(False, point, event, event.from_left_contour)
    event.complement.complement, event.complement = left_event, right_event
    events_queue.push(left_event)
    events_queue.push(right_event)


def bounding_box_in_bounding_box(left: Contour, right: Contour) -> bool:
    ((left_x_min, left_x_max, left_y_min, left_y_max),
     (right_x_min, right_x_max, right_y_min, right_y_max)) = (
        contour_to_bounding_box(left), contour_to_bounding_box(right))
    return (right_x_min <= left_x_min and left_x_max <= right_x_max
            and right_y_min <= left_y_min and left_y_max <= right_y_max)



from enum import (IntEnum,
                  unique)

from robust.angular import (Orientation,
                            orientation)
from robust.linear import segment_contains

from .core import intra_contour as _intra_contour
from .hints import (Contour,
                    Point,
                    Segment)


def point_in_segment(point: Point, segment: Segment) -> bool:
    """
    Checks if point lies inside of the segment or is one of its endpoints.

    Time complexity:
        ``O(1)``
    Memory complexity:
        ``O(1)``

    :param point: point to locate.
    :param segment: segment to check.
    :returns:
        true if point lies inside segment or equal to one of its endpoints,
        false otherwise.

    >>> point_in_segment((0, 0), ((0, 0), (1, 0)))
    True
    >>> point_in_segment((0, 1), ((0, 0), (1, 0)))
    False
    """
    return segment_contains(segment, point)


@unique
class PointContourLocation(IntEnum):
    OUTSIDE = 0
    INSIDE = 1
    BOUNDARY = 2


def point_in_contour(point: Point, contour: Contour) -> PointContourLocation:
    """
    Finds location of point in relation to contour.

    Based on ray casting algorithm.

    Time complexity:
        ``O(len(contour))``
    Memory complexity:
        ``O(1)``
    Reference:
        https://en.wikipedia.org/wiki/Point_in_polygon#Ray_casting_algorithm

    :param point: point to locate.
    :param contour:
        contour to check, vertices should be listed in counterclockwise order.
    :returns: location of point in relation to polygon.

    >>> contour = [(0, 0), (2, 0), (2, 2), (0, 2)]
    >>> point_in_contour((0, 0), contour) is PointContourLocation.BOUNDARY
    True
    >>> point_in_contour((1, 1), contour) is PointContourLocation.INSIDE
    True
    >>> point_in_contour((2, 2), contour) is PointContourLocation.BOUNDARY
    True
    >>> point_in_contour((3, 3), contour) is PointContourLocation.OUTSIDE
    True
    """
    result = False
    _, point_y = point
    for index in range(len(contour)):
        start, end = contour[index - 1], contour[index]
        if point_in_segment(point, (start, end)):
            return PointContourLocation.BOUNDARY
        (_, start_y), (_, end_y) = start, end
        if ((start_y > point_y) is not (end_y > point_y)
                and ((end_y > start_y) is (orientation(end, start, point)
                                           is Orientation.COUNTERCLOCKWISE))):
            result = not result
    return PointContourLocation(result)


def contour_in_contour(left: Contour, right: Contour) -> bool:
    """
    Checks if the contour fully lies inside the region
    bounded by the other one.

    Time complexity:
        ``O((len(left) + len(right)) * log (len(left) + len(right)))``
    Memory complexity:
        ``O((len(left) + len(right)) * log (len(left) + len(right)))``

    :param left:
        contour to check for,
        vertices should be listed in counterclockwise order.
    :param right:
        contour to check in,
        vertices should be listed in counterclockwise order.
    :returns:
        true if the left contour lies inside the other one, false otherwise.

    >>> triangle = [(0, 0), (1, 0), (0, 1)]
    >>> square = [(0, 0), (1, 0), (1, 1), (0, 1)]
    >>> contour_in_contour(triangle, triangle)
    True
    >>> contour_in_contour(triangle, square)
    True
    >>> contour_in_contour(square, triangle)
    False
    >>> contour_in_contour(square, square)
    True
    """
    return (_intra_contour.bounding_box_in_bounding_box(left, right)
            and _intra_contour.sweep(left, right))
